#include <stdio.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <sys/time.h>
/*************************************************************************/
/* Program Name: xnet                                                    */
/* Programmer: Jeff Humphrey (Sat Dec 12 15:16:20 EST 1992)              */
/* Date of Completion (Rev 1.1) (Thu Jan  7 13:54:29 EST 1993)           */
/*                    (Rev 1.2) (Fri Jan  8 18:09:48 EST 1993)           */
/* Purpose: To make a resonable map of the network based on information  */
/*   gathered through remote packet sends with a low (but increasing)    */
/*   time-to-live.  The information will be graphed as window pictures   */
/*   of the network and it's connections from this site.  The data will  */
/*   probably not be terribly useful (just for the sake of information). */
/* Note: Do not use this as an example of good coding style   :-)        */
/* (C) Naval Research Lab, Washington DC                                 */
/* All code written by Jeff Humphrey but owned by NRL.  Not responsible  */
/*   for any bad things that might occur as a result of using, having,   */
/*   or installing this software.  The software is provided AS IS.       */
/* Permission to use and distribute this program is hereby granted as    */
/*   long you aren't charging anything for it.  Use of this source code  */
/*   in other packages or as part of another program cannot happen       */
/*   unless permission is given by the ADP security officer at the Naval */
/*   Research Lab.                                                       */
/* Bug Report: xnet_1_2@hightop.nrl.navy.mil                             */
/*************************************************************************/
#define TRUE                                         1
#define FALSE                                        0
#define PADDING                                    500
#define PORT                              (32768+666+1)
#define MOVING_FALSE                                 0
#define MOVING_BASE                                  1
#define MOVING_NODE                                  2
/*****************************************************/
/* Bits for cursors and the larger 'xnet' image      */
/*****************************************************/
#define SMALL_HOT                                    5
#define SSIZE                                       11
#define LARGE_HOT                                    6
#define LSIZE                                       13
#define POINT_HOT                                    1
#define PSIZE                                       15
#define icon_width                                  53
#define icon_height                                 58
char smask_bits [] = {
  0xff,0x07,0xff,0x07,0xff,0x07,0x07,0x07,0x07,0x07,0x07,0x07,0x07,0x07,0x07,
  0x07,0xff,0x07,0xff,0x07,0xff,0x07};
char ssrc_bits [] = {
  0x00,0x00,0xfe,0x03,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,
  0x02,0x02,0x02,0xfe,0x03,0x00,0x00};
char cmask_bits [] = {
  0xff,0x1f,0xff,0x1f,0xff,0x1f,0x07,0x1c,0x07,0x1c,0x07,0x1c,0x07,0x1c,0x07,
  0x1c,0x07,0x1c,0x07,0x1c,0xff,0x1f,0xff,0x1f,0xff,0x1f};
char csrc_bits [] = {
  0x00,0x00,0xfe,0x0f,0x02,0x08,0x02,0x08,0x02,0x08,0x02,0x08,0x02,0x08,0x02,
  0x08,0x02,0x08,0x02,0x08,0x02,0x08,0xfe,0x0f,0x00,0x00};
char pmask_bits [] = {
  0x03,0x80,0x0f,0x80,0x3e,0x80,0xfe,0x80,0xfc,0x81,0xfc,0x83,0xf8,0x83,0xf8,
  0x81,0xf0,0x83,0x60,0x87,0x00,0x8e,0x00,0x9c,0x00,0xb8,0x00,0x90,0x00,0x80};
char psrc_bits [] = {
  0x00,0x00,0x02,0x00,0x0c,0x00,0x3c,0x00,0xf8,0x00,0xf8,0x01,0x70,0x00,0xb0,
  0x00,0x20,0x01,0x00,0x02,0x00,0x04,0x00,0x08,0x00,0x10,0x00,0x00,0x00,0x00};
char icon_bits [] = {
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0xa0,0x00,0x00,0x00,0xff,0x01,0x00,0xe0,0x10,0x18,0x80,0xff,0x03,0x30,0x78,0x30,
  0x18,0x18,0x38,0x00,0x78,0x38,0x70,0x18,0x1e,0x38,0x00,0xf0,0x1d,0xf0,0x98,0x0f,
  0x38,0x00,0xf0,0x1f,0xe0,0x99,0x07,0x38,0x00,0xe0,0x8f,0xe7,0x9b,0x01,0x38,0x00,
  0xc0,0x8f,0x63,0x9f,0x01,0x38,0x00,0x80,0x0f,0x60,0x9f,0x3f,0x38,0x00,0x80,0x1f,
  0x60,0x9e,0x3f,0x38,0x00,0xc0,0x3f,0x70,0x9c,0x01,0x38,0x00,0xc0,0x79,0x70,0x9c,
  0x03,0x18,0x00,0xe0,0xf0,0x78,0x98,0x0f,0x10,0x00,0x60,0x60,0x78,0x98,0x3f,0x10,
  0x00,0x60,0x00,0x00,0x00,0x7c,0x00,0x00,0x60,0x00,0x00,0x00,0x70,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x03,0xe0,0x00,0x00,0x80,0x01,0xc0,0x01,0x01,0x07,0x00,0x40,0x02,
  0xb0,0x80,0x01,0x08,0x00,0x20,0x04,0x4f,0x40,0x01,0x10,0x00,0xc0,0xff,0x20,0xa0,
  0x00,0x20,0x00,0x00,0x08,0x18,0xd0,0x01,0x40,0x00,0x00,0x10,0x84,0x69,0x01,0x40,
  0x00,0x00,0x00,0x62,0xb5,0x24,0x41,0x00,0x00,0x80,0xc1,0x6f,0xf6,0x61,0x00,0x00,
  0x40,0x3c,0x21,0x4a,0x1f,0x00,0x00,0x20,0x00,0x99,0x81,0x01,0x00,0x00,0x18,0x80,
  0x10,0x00,0x00,0x00,0x00,0x04,0x40,0x10,0x00,0x00,0x00,0x00,0x02,0x20,0x08,0x00,
  0x00,0x00,0x00,0x01,0x20,0x18,0x00,0x00,0x00,0x80,0x00,0x10,0x00,0x00,0x00,0x00,
  0x40,0x00,0x30,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00};
/*****************************************************/
/* Structure Explaination: The following structures  */
/*   (net_type) and (node_type) are used to hold all */
/*   information for a single node in the network.   */
/*   The 'next' field of the node_type is used to    */
/*   point to the next node in the list.  That is to */
/*   say that all nodes in the network are stored in */
/*   a simple linked list of nodes for quick search. */
/*   Also inside of the node type is a 'node' list   */
/*   of type (net_type).  This list is a list of     */
/*   nodes that this particular node is connected    */
/*   back to ... this can and will create loops in   */
/*   the network (and is to be interpreted as a      */
/*   good thing ...)                                 */
/*****************************************************/
typedef struct NL {
  struct N *node;       /* Node entry in list        */
  struct NL *next;      /* Next list item            */
} net_type;
typedef struct N {
  char *name;           /* Name of the node          */
  char *lname;          /* Lowercase name (4 speed)  */
  int x; int y;         /* Location on visual map    */
  net_type *node;       /* Nodes we are connected to */
  struct N *next;       /* Next node in the list     */
} node_type;
typedef struct L {
  char *name;           /* string                    */
  struct L *next;       /* Next item in list         */
} list_type;
int base_x; int base_y;
/*****************************************************/
/* Color Stuff                                       */
/*****************************************************/
#define PURPLE                                                   0
#define ORANGE                                                   1
#define RED                                                      2
#define BLUE                                                     3
#define COLOR_PALE                                               4
#define COLOR_LIGHT                                              5
#define COLOR_MEDIUM                                             6
#define COLOR_DARK                                               7
#define NUMBEROFCOLORS                                           8
typedef struct {
  unsigned long red;
  unsigned long green;
  unsigned long blue;
} color_type;
color_type colors [NUMBEROFCOLORS] = {
  {20000,8328,36480},{65280,42240,0},{65280,17664,0},{0,0,65280},
  {0xfbfb,0xfbfb,0xfbfb},{0xe4e4,0xe4e4,0xe4e4},{0xc7c7,0xc7c7,0xc7c7},
  {0xabab,0xabab,0xabab}
};
unsigned long plane_return [NUMBEROFCOLORS];
unsigned long pixels_return [NUMBEROFCOLORS];
XColor color; XColor clight; XColor cdark;
/*****************************************************/
/* Global Variables                                  */
/*****************************************************/
char *bold_name = "-adobe-courier-bold-r-normal--12-120-75-75-m-70-iso8859-1";
char *bold2_name = "-*-*-bold-r-normal--14-*-*-*-*-*-iso8859-1";
Font font_bold;
XFontStruct *font_struct;
Display *display;
unsigned char sequence;
u_short lastport;
GC gc;
int depth;
int screen;
Window root;
Pixmap ppsrc; Pixmap ppmask;
Pixmap csrc; Pixmap cmask;
Pixmap ssrc; Pixmap smask;
Pixmap helparea;
Cursor pointer;
Cursor mover;
Cursor smover;
Pixmap icon_image;
XImage *image;
XSizeHints hints;
Colormap colormap;
Window window;
Window mapwindow;
Window infowindow;
unsigned long black;
unsigned long white;
XEvent report;
/*****************************************************/
/* Buttons and Windows                               */
/*****************************************************/
#define WINDOW_BASE                                  0
#define NUMBEROFWINDOWS                              1
#define NO_WINDOW                      NUMBEROFWINDOWS
#define TYPE_BOLD                               0x0001
#define TYPE_TEXT                               0x0002
#define INDICATOR                               0x0004
#define TYPE_IMAGE                              0x0008
#define TYPE_DARK                               0x0010
#define LEVEL_0                                 0x1000
#define LEVEL_1                                 0x2000
#define LEVEL_2                                 0x4000
#define LEVEL_3                                 0x8000
#define BOLD          (TYPE_DARK | TYPE_BOLD | LEVEL_2)
#define TEXT          (TYPE_DARK | TYPE_TEXT | LEVEL_1)
#define HIGH_BOLD     (TYPE_DARK | TYPE_BOLD | LEVEL_3)
#define IMAGE                    (TYPE_IMAGE | LEVEL_1)
typedef struct {
  int x; int y; int width; int height; int textx; int texty;
  int flag; char *label;
  int mode; int border; unsigned int bold;
  int max_length;
  char *valid;
  unsigned int label_color;
} button_type;
#define BUTTON_QUIT                                  0
#define BUTTON_ADD                                   1
#define BUTTON_NAME                                  2
#define BUTTON_ERROR                                 3
#define BUTTON_ACTIVE                                4
#define BUTTON_SHOW                                  5
#define BUTTON_INFO                                  6
#define NUMBEROFBUTTONS                              7
#define NO_BUTTON                      NUMBEROFBUTTONS
button_type Buttons [NUMBEROFBUTTONS] = {
  16,15,45,16,8,12,FALSE,"Quit",WINDOW_BASE,FALSE,HIGH_BOLD,0,"",PURPLE,
  395,15,53,16,8,12,FALSE,"Trace",WINDOW_BASE,FALSE,BOLD,0,"",PURPLE,
  88,15,280,16,7,12,FALSE,"net.net.net.net.net.net.net.net.NET.ne",
    WINDOW_BASE,FALSE,TEXT,38,"",PURPLE,
  88,40,280,16,7,12,FALSE,"                                      ",
    WINDOW_BASE,FALSE,TEXT,38,"",PURPLE,
  467,15,16,16,8,12,FALSE,"",WINDOW_BASE,FALSE,LEVEL_1,0,"",PURPLE,
  88,62,280,16,7,12,FALSE,"                                      ",
    WINDOW_BASE,FALSE,TEXT,38,"",RED,
  12,40,54,59,1,1,FALSE,"",WINDOW_BASE,FALSE,IMAGE,0,"",PURPLE,
};
/*****************************************************/
/* Stuff for sockets and workings of the code        */
/*****************************************************/
int flag;
typedef struct {
  struct ip ip;
  struct udphdr header;
  u_char seq;
  u_char ttl;
} packet_type;
typedef struct {
  int size;
  u_char packet [512];
} packet_return_type;
packet_return_type packet;
packet_type *packet_out;
unsigned int pad_size = PADDING;
unsigned int packet_size;
int sock; int channel_out;
struct protoent *channel_proto;
struct sockaddr_in source;
struct sockaddr_in *destination;
struct sockaddr whereto;
struct hostent *host;
int time_to_live;
char ret_name [1000];
list_type *list;
node_type *network;     /* The complete network map  */
int current_x;          /* X for added nodes         */
int current_y;          /* Y for added nodes         */
char *lines [] = {
  "Name: xnet (X-Net version 1.2)",
  "Programmer: Jeff Humphrey (Locus Inc)",
  "Date: Fri Jan  8 18:09:48 EST 1993",
  "Copyright: Created by the Computer Security Office",
  "  at the Naval Research Lab, Washington DC.  The",
  "  program is distributed as public domain software.",
  "  The author claims no responsibility for the codes",
  "  use or effect on equipment.  The software is",
  "  provided AS IS.",
  "Credits: Based on ideas presented by Van Jacobson in",
  "  his program (traceroute).  Special thanks to Les",
  "  Aker and Richard Deal for evaluating xnet ...",
  "Bug Report: E-mail xnet_1_2@hightop.nrl.navy.mil",
  0
};
int lowercase (string)
char *string;
/*****************************************************/
/* Function Name: lowercase                          */
/* Purpose: Converts string to lowercase             */
/*****************************************************/
{
  int length;
  int counter;
  length = strlen (string);
  for (counter = 0;counter < length;counter++)
    if ((string [counter] >= 'A') && (string [counter] <= 'Z'))
      string [counter] = string [counter] - 'A' + 'a';
  return (TRUE);
}
int three_d_up (x,y,width,height,level)
unsigned int x;
unsigned int y;
unsigned int width;
unsigned int height;
unsigned int level;
/*****************************************************/
/* Function Name: three_d_up                         */
/*****************************************************/
{
  int counter;
  if (depth == 8)
    {
      for (counter = 0;counter < level;counter++)
        {
          XSetForeground (display,gc,pixels_return [COLOR_PALE]);
          XDrawLine (display,window,gc,x - counter,y - counter,
                                       x - counter,y + height + counter);
          XDrawLine (display,window,gc,x - counter,y - counter,
                                       x + width + counter,y - counter);
          XSetForeground (display,gc,pixels_return [COLOR_DARK]);
          XDrawLine (display,window,gc,x + width + counter,y - counter,
                     x + width + counter,y + height + counter);
          XDrawLine (display,window,gc,x - counter,y + height + counter,
                     x + width + counter,y + height + counter);
        }
    }
  else
    {
      XSetForeground (display,gc,black);
      XDrawRectangle (display,window,gc,x,y,width,height);
    }
  XFlush (display);
  return (FALSE);
}
unsigned int button_press (button,event)
button_type button;
XEvent event;
/*****************************************************/
/* Function Name: button_press                       */
/* Purpose: To check and see if any buttons were     */
/*   pressed.  If yes ... return TRUE ... if no then */
/*   return FALSE.  This function only allows a      */
/*   button to be on the base window 'window'.       */
/*****************************************************/
{
  if (button.mode != NO_WINDOW)
    switch (event.type) {
      case (ButtonPress): case (ButtonRelease):
        return ((event.xbutton.x > (button.x - 2)) &&
                (event.xbutton.x < (button.x + button.width + 4)) &&
                (event.xbutton.y > (button.y - 2)) &&
                (event.xbutton.y < (button.y + button.height + 4)) &&
                (event.xbutton.window == window));
        break;
      default: break;
    } else return (FALSE);
}
unsigned int any_buttons (event)
XEvent event;
/*****************************************************/
/* Function Name: any_buttons                        */
/* Purpose: To check and see if any buttons were     */
/*   pressed ... return button -OR- NO_BUTTON ...    */
/*****************************************************/
{
  int counter;
  unsigned int button;
  button = NO_BUTTON;
  for (counter = 0;counter < NUMBEROFBUTTONS;counter++)
    if (button_press (Buttons [counter],event) == TRUE)
      button = counter;
  return (button);
}
int drawing (button)
button_type button;
/*****************************************************/
/* Function Name: drawing                            */
/*****************************************************/
{
  if (depth == 8)
    {
      if ((button.bold | TYPE_DARK) == button.bold)
        {
          if (button.flag == TRUE)
            XSetForeground (display,gc,pixels_return [COLOR_MEDIUM]);
          else
            XSetForeground (display,gc,pixels_return [COLOR_LIGHT]);
          XFillRectangle (display,window,gc,button.x,button.y,button.width + 1,
                          button.height + 1);
        }
      if (button.flag == TRUE)
        XSetForeground (display,gc,pixels_return [COLOR_DARK]);
      else
        XSetForeground (display,gc,pixels_return [COLOR_PALE]);
      if (((button.bold | LEVEL_1) == button.bold) ||
          ((button.bold | LEVEL_2) == button.bold) ||
          ((button.bold | LEVEL_3) == button.bold))
        {
          XDrawLine (display,window,gc,button.x - 1,button.y - 1,
                                       button.x + button.width + 1,button.y - 1);
          XDrawLine (display,window,gc,button.x - 1,button.y - 1,
                                       button.x - 1,button.y + button.height + 1);
        }
      if (((button.bold | LEVEL_2) == button.bold) ||
          ((button.bold | LEVEL_3) == button.bold))
        {
          XDrawLine (display,window,gc,button.x - 2,button.y - 2,
                                       button.x + button.width + 2,button.y - 2);
          XDrawLine (display,window,gc,button.x - 2,button.y - 2,
                                 button.x - 2,button.y + button.height + 2);
        }
      if ((button.bold | LEVEL_3) == button.bold)
        {
          XDrawLine (display,window,gc,button.x - 3,button.y - 3,
                                       button.x + button.width + 3,button.y - 3);
          XDrawLine (display,window,gc,button.x - 3,button.y - 3,
                                 button.x - 3,button.y + button.height + 3);
        }
      if (button.flag == TRUE)
        XSetForeground (display,gc,pixels_return [COLOR_PALE]);
      else
        XSetForeground (display,gc,pixels_return [COLOR_DARK]);
      if (((button.bold | LEVEL_1) == button.bold) ||
          ((button.bold | LEVEL_2) == button.bold) ||
          ((button.bold | LEVEL_3) == button.bold))
        {
          XDrawLine (display,window,gc,button.x,button.y + button.height + 1,
                             button.x + button.width + 1,button.y + button.height + 1);
          XDrawLine (display,window,gc,button.x + button.width + 1,
                             button.y,button.x + button.width + 1,
                             button.y + button.height + 1);
        }
      if (((button.bold | LEVEL_2) == button.bold) ||
          ((button.bold | LEVEL_3) == button.bold))
        {
          XDrawLine (display,window,gc,button.x - 1,
                             button.y + button.height + 2,
                             button.x + button.width + 2,
                             button.y + button.height + 2);
          XDrawLine (display,window,gc,button.x + button.width + 2,
                                 button.y - 1,button.x + button.width + 2,
                                 button.y + button.height + 2);
        }
      if ((button.bold | LEVEL_3) == button.bold)
        {
          XDrawLine (display,window,gc,button.x - 2,
                             button.y + button.height + 3,
                             button.x + button.width + 3,
                             button.y + button.height + 3);
          XDrawLine (display,window,gc,button.x + button.width + 3,
                             button.y - 2,button.x + button.width + 3,
                             button.y + button.height + 3);
        }
      XSetForeground (display,gc,black);
      if (((button.bold | TYPE_BOLD) == button.bold) ||
          ((button.bold | TYPE_TEXT) == button.bold))
        {
          XSetForeground (display,gc,pixels_return [button.label_color]);
          XDrawString (display,window,gc,button.textx + button.x,
                       button.texty + button.y,button.label,strlen (button.label));
        }
      if ((button.bold | TYPE_IMAGE) == button.bold)
        XPutImage (display,window,gc,image,0,0,button.x + 1,button.y + 1,
                                           icon_width,icon_height);
    }
  else
    {
      if (button.flag == FALSE)
        XSetForeground (display,gc,white);
      else
        XSetForeground (display,gc,black);
      XFillRectangle (display,window,gc,button.x,button.y,button.width + 1,
                      button.height + 1);
      XSetForeground (display,gc,black);
      XDrawRectangle (display,window,gc,button.x,button.y,button.width + 1,
                      button.height + 1);
      if (((button.bold | TYPE_BOLD) == button.bold) ||
          ((button.bold | TYPE_TEXT) == button.bold))
        {
          if (button.flag == FALSE)
            XSetForeground (display,gc,black);
          else
            XSetForeground (display,gc,white);
          XDrawString (display,window,gc,button.textx + button.x,
                       button.texty + button.y,button.label,strlen (button.label));
        }
      if ((button.bold | TYPE_IMAGE) == button.bold)
        XPutImage (display,window,gc,image,0,0,button.x + 1,button.y + 1,
                                           icon_width,icon_height);
    }
  XFlush (display);
  return (FALSE);
}
int write_error (message)
char *message;
/*****************************************************/
/* Function Name: write_error                        */
/* Purpose: The purpose of this function is to write */
/*   out an error message to the error line on the   */
/*   screen.                                         */
/*****************************************************/
{
  strcpy (Buttons [BUTTON_ERROR].label,message);
  drawing (Buttons [BUTTON_ERROR]);
}
int stop_trace ()
/*****************************************************/
/* Function Name: stop_trace                         */
/* Purpose: The purpose of this function is to have  */
/*   a convenient way to stop tracing.               */
/*****************************************************/
{
  lastport = 0xffff;
  if (Buttons [BUTTON_ADD].flag == TRUE)
    {
      Buttons [BUTTON_ADD].flag = FALSE;
      drawing (Buttons [BUTTON_ADD]);
      if (Buttons [BUTTON_ACTIVE].flag == TRUE)
        {
          Buttons [BUTTON_ACTIVE].flag = FALSE;
          drawing (Buttons [BUTTON_ACTIVE]);
        }
      lastport = 0xffff;
      list = NULL;
      write_error ("Stopping Trace");
    }
}
int update_window (win)
Window win;
/*****************************************************/
/* Function Name: update_window                      */
/*****************************************************/
{
  int counter;
  char tempstr [100];
  if (win == window)
    {
      three_d_up (5,5,67,100,1);
      three_d_up (78,5,300,100,1);
      three_d_up (384,5,110,100,1);
      for (counter = 0;counter < NUMBEROFBUTTONS;counter++)
        if (Buttons [counter].mode == WINDOW_BASE)
          drawing (Buttons [counter]);
    }
  if (win == mapwindow)
    {
      node_type *temp;
      draw_node ((node_type *) NULL,pixels_return [BLUE]);
      for (temp = network;temp != (node_type *) NULL;temp = temp->next)
        draw_node (temp,black);
    }
  if (win == infowindow)
    {
      XSetForeground (display,gc,pixels_return [RED]);
      XFillRectangle (display,infowindow,gc,5,5,400 - 10,10);
      XFillRectangle (display,infowindow,gc,5,260 - 15,400 - 10,10);
      XSetForeground (display,gc,black);
      for (counter = 0;lines [counter] != 0;counter++)
        XDrawString (display,infowindow,gc,10,(counter * 15) + 43,
                     lines [counter],strlen (lines [counter]));
    }
  XFlush (display);
  return (TRUE);
}
int toggle_button (button)
unsigned int button;
/*****************************************************/
/* Function Name: toggle_button                      */
/*****************************************************/
{
  Buttons [button].flag = TRUE - Buttons [button].flag;
  drawing (Buttons [button]);
}
int check_button (button)
unsigned int button;
/*****************************************************/
/* Function Name: check_button                       */
/*****************************************************/
{
  switch (button) {
    case (BUTTON_QUIT):
      toggle_button (BUTTON_QUIT);
      break;
    case (BUTTON_ADD):
      if (Buttons [BUTTON_ADD].flag == TRUE)
        { lastport = 0xffff; stop_trace (); }
      else
        { toggle_button (BUTTON_ADD); trace (); }
      break;
    case (BUTTON_NAME):
      stop_trace ();
      strcpy (Buttons [BUTTON_NAME].label,"");
      drawing (Buttons [BUTTON_NAME]);
      break;
    case (BUTTON_INFO):
      toggle_button (BUTTON_INFO);
      if (Buttons [BUTTON_INFO].flag == TRUE)
        XMapWindow (display,infowindow);
      else
        XUnmapWindow (display,infowindow);
      XFlush (display);
      break;
    default: break;
  }
  return (TRUE);
}
int open_display (ac,av)
int ac;
char **av;
/*****************************************************/
/* Function Name: open_display                       */
/* Purpose: The purpose of this function is simply   */
/*   to open everything we need for the display.     */
/*****************************************************/
{
  int counter;
  XColor one;
  XColor two;
  char hostname [100];
  if ((display = XOpenDisplay ("")) == (Display *) NULL)
    {
      if ((display = XOpenDisplay ("unix:0")) == (Display *) NULL)
        {
          gethostname (hostname,sizeof (hostname));
          strcat (hostname,":0");
          if ((display = XOpenDisplay (hostname)) == (Display *) NULL)
            {
              printf ("Can't Open ()\n");
              printf ("           (unix:0)\n");
              printf ("      -OR- (%s)\n",hostname);
              exit (1);
            }
        }
    }
  screen = DefaultScreen (display);
  depth = DefaultDepth (display,screen);
  colormap = DefaultColormap (display,screen);
  black = BlackPixel (display,screen);
  if (depth == 8)
    {
      XAllocColorCells (display,colormap,TRUE,plane_return,0,
                        pixels_return,NUMBEROFCOLORS);
      for (counter = 0;counter < NUMBEROFCOLORS;counter++)
        {
          color.pixel = pixels_return [counter];
          color.red = colors [counter].red;
          color.green = colors [counter].green;
          color.blue = colors [counter].blue;
          color.flags = DoRed | DoGreen | DoBlue;
          XStoreColor (display,colormap,&color);
        }
      white = pixels_return [COLOR_LIGHT];
    }
  else
    {
      for (counter = 0;counter < NUMBEROFCOLORS;counter++)
        pixels_return [counter] = black;
      white = WhitePixel (display,screen);
      pixels_return [COLOR_PALE] = white;
      pixels_return [COLOR_LIGHT] = white;
      pixels_return [COLOR_MEDIUM] = white;
    }
  gc = DefaultGC (display,screen);
  font_struct = XLoadQueryFont (display,bold_name);
  if (font_struct == (XFontStruct *) NULL)
    {
      font_struct = XLoadQueryFont (display,bold2_name);
      if (font_struct == (XFontStruct *) NULL)
        { printf ("Couldn't load any of my fonts ?\n"); exit (1); }
      else
        font_bold = font_struct->fid;
    }
  else
    font_bold = font_struct->fid;
  XSetFont (display,gc,font_bold);
  root = RootWindow (display,screen);
  window = XCreateSimpleWindow (display,root,10,10,500,500,3,black,white);
  mapwindow = XCreateSimpleWindow (display,window,5,111,488,382,1,
                         pixels_return [COLOR_DARK],pixels_return [COLOR_MEDIUM]);
  infowindow = XCreateSimpleWindow (display,window,60,90,400,260,1,
                         pixels_return [COLOR_DARK],pixels_return [COLOR_LIGHT]);
  helparea = XCreatePixmap (display,window,500,500,depth);
  if (depth == 8)
    icon_image = XCreatePixmapFromBitmapData (display,window,icon_bits,icon_width,
                 icon_height,pixels_return [ORANGE],pixels_return [COLOR_LIGHT],depth);
  else
    icon_image = XCreatePixmapFromBitmapData (display,window,icon_bits,icon_width,
                 icon_height,black,white,depth);
  ppsrc = XCreatePixmapFromBitmapData (display,window,psrc_bits,
                          PSIZE,PSIZE,0xffff,0x0000,1);
  ppmask = XCreatePixmapFromBitmapData (display,window,pmask_bits,
                          PSIZE,PSIZE,0xffff,0x0000,1);
  csrc = XCreatePixmapFromBitmapData (display,window,csrc_bits,
                          LSIZE,LSIZE,0xffff,0x0000,1);
  cmask = XCreatePixmapFromBitmapData (display,window,cmask_bits,
                          LSIZE,LSIZE,0xffff,0x0000,1);
  ssrc = XCreatePixmapFromBitmapData (display,window,ssrc_bits,
                          SSIZE,SSIZE,0xffff,0x0000,1);
  smask = XCreatePixmapFromBitmapData (display,window,smask_bits,
                          SSIZE,SSIZE,0xffff,0x0000,1);
  one.pixel = black;
  two.pixel = WhitePixel (display,screen);
  XQueryColor (display,colormap,&one);
  XQueryColor (display,colormap,&two);
  pointer = XCreatePixmapCursor (display,ppsrc,ppmask,&one,&two,POINT_HOT,POINT_HOT);
  two.pixel = pixels_return [COLOR_MEDIUM];
  XQueryColor (display,colormap,&two);
  smover = XCreatePixmapCursor (display,ssrc,smask,&one,&two,SMALL_HOT,SMALL_HOT);
  one.pixel = pixels_return [BLUE];
  XQueryColor (display,colormap,&one);
  mover = XCreatePixmapCursor (display,csrc,cmask,&one,&two,LARGE_HOT,LARGE_HOT);
  XDefineCursor (display,window,pointer);
  XDefineCursor (display,mapwindow,pointer);
  XDefineCursor (display,infowindow,pointer);
  image = XGetImage (display,icon_image,0,0,icon_width,icon_height,AllPlanes,XYPixmap);
  hints.x = 0; hints.y = 0; hints.width = 500; hints.height = 500;
  hints.min_width = 500; hints.max_width = 500;
  hints.min_height = 500; hints.max_height = 500;
  hints.flags = (PPosition | PSize | PMinSize | PMaxSize);
  icon_image = XCreatePixmapFromBitmapData (display,window,icon_bits,icon_width,
                 icon_height,0,1,1);
  XSetStandardProperties (display,window,"XNET","XNET",icon_image,
                          av,ac,&hints);
  XSelectInput (display,window,(ButtonPressMask | StructureNotifyMask |
                            ExposureMask | KeyPressMask));
  XSelectInput (display,mapwindow,(ButtonPressMask | StructureNotifyMask |
                            ExposureMask | PointerMotionMask |
                            LeaveWindowMask | ButtonReleaseMask));
  XSelectInput (display,infowindow,(ButtonPressMask | StructureNotifyMask |
                            ExposureMask));
  XStoreName (display,window," X-Net (1.2)");
  XMapWindow (display,window);
  XMapWindow (display,mapwindow);
  XFlush (display);
}
int highlight_node (node)
node_type *node;
/*****************************************************/
/* Function Name: highlight_node                     */
/* Purpose: Turn the little node into a red or blue  */
/*   square for movement purposes.                   */
/*****************************************************/
{
  if (node == (node_type *) NULL)
    {
      XSetForeground (display,gc,pixels_return [COLOR_MEDIUM]);
      XFillRectangle (display,mapwindow,gc,base_x - 5,base_y - 5,11,11);
      XSetForeground (display,gc,pixels_return [RED]);
      XDrawRectangle (display,mapwindow,gc,base_x - 5,base_y - 5,10,10);
    }
  else
    {
      XSetForeground (display,gc,pixels_return [COLOR_MEDIUM]);
      XFillRectangle (display,mapwindow,gc,node->x - 4,node->y - 4,9,9);
      XSetForeground (display,gc,pixels_return [RED]);
      XDrawRectangle (display,mapwindow,gc,node->x - 4,node->y - 4,8,8);
    }
}
int draw_connects (node,thiscolor)
node_type *node;
unsigned long thiscolor;
/*****************************************************/
/* Function Name: draw_connects                      */
/* Purpose: Draw all connections ... TO ... the node */
/*   provided by the function (not from this node to */
/*   the one(s) it is connected too)                 */
/*****************************************************/
{
  node_type *temp;
  node_type *temper;
  net_type *net;
  net_type *ttnet;
  for (temp = network;temp != (node_type *) NULL;temp = temp->next)
    {
      if (thiscolor == pixels_return [COLOR_MEDIUM])
        XSetForeground (display,gc,pixels_return [COLOR_MEDIUM]);
      else
        XSetForeground (display,gc,pixels_return [RED]);
      if (temp->node != (net_type *) NULL)
        {
          for (net = temp->node;net != (net_type *) NULL;net = net->next)
            {
              if ((net->node == (node_type *) NULL) && (node == (node_type *) NULL))
                XDrawLine (display,mapwindow,gc,temp->x,temp->y,base_x,base_y);
              else
                {
                  temper = net->node;
                  if (temper == node)
                    XDrawLine (display,mapwindow,gc,temp->x,temp->y,node->x,node->y);
                }
            }
        }
    }
  XFlush (display);
}
int move_node (node,x,y)
node_type *node;
int x; int y;
/*****************************************************/
/* Function Name: move_node                          */
/* Purpose: Move the node from one location on the   */
/*   screen to another one ...                       */
/*****************************************************/
{
  if ( ((x >= 10) && (x <= (488 - 10))) &&
       ((y >= 10) && (y <= (382 - 10))))
    {
      draw_connects (node,pixels_return [COLOR_MEDIUM]);
      draw_node (node,pixels_return [COLOR_MEDIUM]);
      if (node == (node_type *) NULL)
        { base_x = x; base_y = y; }
      else
        { node->x = x; node->y = y; }
      draw_connects (node,black);
      draw_node (node,pixels_return [BLUE]);
    }
  else
    draw_node (node,pixels_return [BLUE]);
  return (TRUE);
}
int check_key (report,button)
XEvent report;
unsigned int button;
/*****************************************************/
/* Function Name: check_key                          */
/*****************************************************/
{
  unsigned int count; static KeySym keysym;
  static XComposeStatus compose;
  static char buffer [100];
  static int bufsize = 100;
  unsigned int length;

  count = XLookupString ( (XKeyEvent *)&report, buffer, bufsize, &keysym, &compose);
  if (keysym == XK_Return)
    return (TRUE);
  else
    if ((keysym != XK_Shift_L) || (keysym != XK_Shift_R))
      {
        if (((keysym == XK_BackSpace) || (keysym == XK_Delete)) &&
            ((length = strlen (Buttons [button].label)) > 0))
          {
            Buttons [button].label [length - 1] = NULL;
            XSetForeground (display,gc,WhitePixel (display,screen));
            XFillRectangle (display,window,gc,Buttons [button].x + 2,
                            Buttons [button].y + 2,
                            Buttons [button].width - 3,Buttons [button].height - 3);
            XSetForeground (display,gc,BlackPixel (display,screen));
          }
        else
          if (((keysym >= XK_a) && (keysym <= XK_z)) ||
              ((keysym >= XK_A) && (keysym <= XK_Z)) ||
              ((keysym >= XK_0) && (keysym <= XK_9)) ||
              (keysym == XK_plus) ||
              (keysym == XK_minus) ||
              (keysym == XK_period) ||
              (keysym == XK_colon) ||
              (keysym == XK_underscore))
            {
              if ((strlen (Buttons [button].label) + (strlen (buffer))
                   < Buttons [button].max_length))
                strcat (Buttons [button].label,buffer);
            }
      }
  drawing (Buttons [button]);
  return (FALSE);
}
int fire_packet (ttl)
int ttl;
/*****************************************************/
/* Function Name: fire_packet                        */
/*****************************************************/
{
  struct ip *ip;
  struct udphdr *header;
  int number;
  ip = &packet_out->ip;
  header = &packet_out->header;
  ip->ip_off = 0;
  ip->ip_p = IPPROTO_UDP;
  ip->ip_len = packet_size;
  ip->ip_ttl = ttl;
  ip->ip_dst = destination->sin_addr;
  ip->ip_tos = 0;
  ip->ip_src = source.sin_addr;
  header->uh_sum = 0;
  header->uh_dport = PORT + sequence;
  lastport = header->uh_dport;
  header->uh_sport = (getpid() & 0xffff) | 0x8000;;
  header->uh_ulen = htons ((u_short) (packet_size - sizeof (struct ip)));
  packet_out->ttl = ttl;
  packet_out->seq = sequence;
  if (sequence < 0xaa) sequence++; else sequence = 0x00;
  number = sendto (channel_out,(char *)packet_out,packet_size,0,&whereto,
                   sizeof (struct sockaddr));
  Buttons [BUTTON_ACTIVE].flag = TRUE;
  return (TRUE);
}
int type_of_packet ()
/*****************************************************/
/* Function Name: type_of_packet                     */
/*****************************************************/
{
  struct icmp *icp; int good;
  struct ip *ip;
  unsigned long ditto;
  struct udphdr *rup;
  struct hostent *myhost;
  char name [100];
  int length; int size;
  ip = (struct ip *) packet.packet;
  length = ip->ip_hl << 2;
  size = packet.size - length;
  icp = (struct icmp *) (packet.packet + length);
  rup = (struct udphdr *) (packet.packet + length + sizeof (struct icmp));
  if (rup->uh_dport == lastport)
    {
      good = FALSE;
      strcpy (name,inet_ntoa (ip->ip_src));
      myhost = gethostbyaddr ((const char *)&ip->ip_src,sizeof (ip->ip_src),AF_INET);
      if (myhost != (struct hostent *) NULL)
        {
          if (strlen (myhost->h_name) >= 2)
            { good = TRUE; strcpy (ret_name,myhost->h_name); }
        }
      else
        {
          if (strlen (name) >= 2)
            { good = TRUE; strcpy (ret_name,name); }
        }
      if (good == FALSE)
        strcpy (ret_name,"");
      if ((icp->icmp_code == ICMP_UNREACH_PORT) && (icp->icmp_type == ICMP_UNREACH))
        return (TRUE); else return (FALSE);
    }
  else
    return (FALSE);
}
int catch_packet ()
/*****************************************************/
/* Function Name: catch packet                       */
/*****************************************************/
{
  fd_set fds;
  struct timeval wait;
  int cc = 0;
  int source_size = sizeof (struct sockaddr_in);
  FD_ZERO (&fds);
  FD_SET (sock,&fds);
  wait.tv_sec =  0;
  wait.tv_usec = 2;
  if (select (sock+1,&fds,(fd_set *)0,(fd_set *)0,&wait) > 0)
    cc = recvfrom (sock,(char *) packet.packet,sizeof (packet.packet),0,
                   (struct sockaddr *) &source,&source_size);
  if (cc != 0)
    {
      packet.size = cc;
      return (TRUE);
    }
  return (FALSE);
}
int trace ()
/*****************************************************/
/* Function Name: trace                              */
/* Purpose: The purpose of this function is to start */
/*   a trace and fire off a single packet to the     */
/*   destination machine.  It is up to some other    */
/*   function to watch for the results of the fire.  */
/*   Sockets should already be open and waiting for  */
/*   this ...                                        */
/*****************************************************/
{
struct sockaddr_in * destination;
  char *hostname;
  hostname = Buttons [BUTTON_NAME].label;
  if (strlen (Buttons [BUTTON_NAME].label) == 0)
    {
      write_error ("Hostname Required");
      Buttons [BUTTON_ADD].flag = FALSE;
      lastport = 0xffff;
      drawing (Buttons [BUTTON_ADD]);
      return (FALSE);
    }
  destination = (struct sockaddr_in *) &whereto;

  (void) bzero ((char *) &whereto,sizeof (struct sockaddr));
  destination->sin_family = AF_INET;
  if ((destination->sin_addr.s_addr = inet_addr (hostname)) == -1)
    {
      if ((host = gethostbyname (hostname)) != (struct hostent *) NULL)
        {
          destination->sin_family = host->h_addrtype;
          bcopy (host->h_addr,(caddr_t)&destination->sin_addr,host->h_length);
          hostname = host->h_name;
        }
      else
        {
          write_error ("Hostname Unknown");
          Buttons [BUTTON_ADD].flag = FALSE;
          lastport = 0xffff;
          drawing (Buttons [BUTTON_ADD]);
          return (FALSE);
        }
    }
  time_to_live = 1;
  list = (list_type *) NULL;
  for (;catch_packet () == TRUE;);
  write_error ("Tracing Connection");
  return (TRUE);
}
int draw_node (node,thiscolor)
node_type *node;
unsigned long thiscolor;
/*****************************************************/
/* Function Name: draw_node                          */
/*****************************************************/
{
  node_type *temp;
  node_type *temper;
  net_type *net;
  net_type *tnet;
  if (node == (node_type *) NULL)
    {
      XSetForeground (display,gc,thiscolor);
      XFillRectangle (display,mapwindow,gc,base_x - 5,base_y - 5,11,11);
      XFlush (display);
      return (TRUE);
    }
  if (thiscolor == pixels_return [COLOR_MEDIUM])
    XSetForeground (display,gc,pixels_return [COLOR_MEDIUM]);
  else
    XSetForeground (display,gc,pixels_return [RED]);
  if (node->node == (net_type *) NULL)
    XDrawLine (display,mapwindow,gc,node->x,node->y,base_x,base_y);
  else
    {
      for (tnet = node->node;tnet != (net_type *) NULL;tnet = tnet->next)
        {
          temper = tnet->node;
          if (temper == (node_type *) NULL)
            XDrawLine (display,mapwindow,gc,node->x,node->y,base_x,base_y);
          else
            XDrawLine (display,mapwindow,gc,node->x,node->y,temper->x,temper->y);
        }
    }
  if (thiscolor == pixels_return [COLOR_MEDIUM])
    XSetForeground (display,gc,pixels_return [COLOR_MEDIUM]);
  else
    XSetForeground (display,gc,black);
  XFillRectangle (display,mapwindow,gc,node->x - 4,node->y - 4,9,9);
  XFlush (display);
  return (TRUE);
}
node_type *find_node_by_name (name)
char *name;
/*****************************************************/
/* Function Name: find_node_by_name                  */
/* Purpose: If the named node is present in the      */
/*   network map ... this function will return it.   */
/*   If not .. it'll return (node_type *) NULL.      */
/*****************************************************/
{
  node_type *temp;
  char *lower;
  int counter;
  lower = (char *) malloc (strlen (name));
  if (lower == (char *) NULL)
    { printf ("malloc: NULL in find_node_by_name\n"); exit (0); }
  strcpy (lower,name);
  lowercase (lower);
  for (temp = network;temp != (node_type *) NULL;temp = temp->next)
    if (strcmp (temp->lname,lower) == 0)
      {
        free ((char *) lower);
        return ((node_type *) temp);
      }
  free ((char *) lower);
  return ((node_type *) NULL);
}
int node_to_network (name,connected)
char *name;
char *connected;
/*****************************************************/
/* Function Name: node_to_network                    */
/*****************************************************/
{
  node_type *temp;
  node_type *temper;
  node_type *contact;
  node_type *further;
  net_type *tnet; net_type *rnet; net_type *snet;
  net_type *net;
  char lower [100]; char lower2 [100];
  strcpy (lower,name);
  lowercase (lower);
  contact = (node_type *) NULL;
  for (temp = network;temp != (node_type *) NULL;temp = temp->next)
    if (strcmp (temp->lname,lower) == 0)
      contact = temp;
  if (contact == (node_type *) NULL)
    {
      if (network == (node_type *) NULL)
        {
          temp = (node_type *) malloc (sizeof (node_type));
          temp->next = (node_type *) NULL;
          temp->name = (char *) malloc (strlen (name));
          strcpy (temp->name,name);
          temp->lname = (char *) malloc (strlen (name));
          strcpy (temp->lname,name);
          lowercase (temp->lname);
          temp->x = current_x; temp->y = current_y;
          current_x += 20;
          if (current_x > 300)
            {
              current_x = 20;
              current_y += 20;
              if (current_y > 100)
                current_y = 20;
            }
          temp->node = (net_type *) malloc (sizeof (net_type));
          net = temp->node;
          net->node = (node_type *) NULL;
          net->next = (net_type *) NULL;
          draw_node (temp,black);
          network = temp;
        }
      else
        {
          if (strlen (name) > 0)
            {
              contact = find_node_by_name (connected);
              for (temp = network;temp->next != (node_type *) NULL;temp = temp->next);
              temp->next = (node_type *) malloc (sizeof (node_type));
              temp = temp->next;
              temp->next = (node_type *) NULL;
              temp->name = (char *) malloc (strlen (name));
              strcpy (temp->name,name);
              temp->lname = (char *) malloc (strlen (name));
              strcpy (temp->lname,name);
              lowercase (temp->lname);
              temp->x = current_x; temp->y = current_y;
              current_x += 20;
              if (current_x > 300)
                {
                  current_x = 20;
                  current_y += 20;
                  if (current_y > 100)
                    current_y = 20;
                }
              temp->node = (net_type *) malloc (sizeof (net_type));
              net = temp->node;
              net->node = (node_type *) contact;
              net->next = (net_type *) NULL;
              draw_node (temp,black);
            }
        }
    }
  else
    {
      further = find_node_by_name (connected);
      if (contact->node != (net_type *) NULL)
        {
          net = contact->node;
          if (net->node != (node_type *) NULL)
            {
              /*** Is it in the list ***/
              strcpy (lower2,connected);
              lowercase (lower2);
              snet = (net_type *) NULL;
              for (rnet = net;rnet != (net_type *) NULL;rnet = rnet->next)
                {
                  temp = rnet->node;
                  if (strcmp (temp->lname,lower2) == 0)
                    snet = rnet;
                }
              if (snet == (net_type *) NULL)
                {
                  /*** Add to list ***/
                  tnet = (net_type *) malloc (sizeof (net_type));
                  tnet->node = further;
                  tnet->next = contact->node;
                  contact->node = tnet;
                }
            }
          draw_node (contact,black);
        }
    }
}
int add_node (name)
char *name;
/*****************************************************/
/* Function Name: add_node                           */
/*****************************************************/
{
  list_type *temp;
  char connected [200];
  if (list == (list_type *) NULL)
    {
      list = (list_type *) malloc (sizeof (list_type));
      list->next = (list_type *) NULL;
      list->name = (char *) malloc (strlen (ret_name));
      strcpy (list->name,ret_name);
      strcpy (connected,"");
    }
  else
    {
      for (temp = list;temp->next != (list_type *) NULL;temp = temp->next);
      strcpy (connected,temp->name);
      temp->next = (list_type *) malloc (sizeof (list_type));
      temp = temp->next;
      temp->next = (list_type *) NULL;
      temp->name = (char *) malloc (strlen (ret_name));
      strcpy (temp->name,ret_name);
    }
  node_to_network (name,connected);
  return (TRUE);
}
node_type *touching_node (x,y)
int x; int y;
/*****************************************************/
/* Function Name: touching_node                      */
/* Purpose: This function will return the node being */
/*   'touched' by the (x,y) if there is one .. else  */
/*   returns ((node_type *) NULL) ...                */
/*****************************************************/
{
  node_type *temp;
  for (temp = network;temp != (node_type *) NULL;temp = temp->next)
    if (((x >= temp->x - 5) && (x <= temp->x + 5)) &&
        ((y >= temp->y - 5) && (y <= temp->y + 5)))
      return (temp);
  return ((node_type *) NULL);
}
int main (argc,argv)
int argc;
char **argv;
/*************************************************************************/
/* Program Name: xnet                                                    */
/* Programmer: Jeff Humphrey (Sat Dec 12 15:16:20 EST 1992)              */
/*************************************************************************/
{
  list_type *temp;
  unsigned int button;
  int counter; int idle;
  int moving_machine;
  node_type *node_move;
  node_type *tempnode;
  XEvent tempe;
  if (geteuid () != 0)
    { printf ("Program must be suid root\n"); exit (0); }
  if ((channel_proto = getprotobyname ("icmp")) == (struct protoent *) NULL)
    { printf ("unknown: icmp\n"); exit (0); }
  if ((sock = socket (AF_INET,SOCK_RAW,channel_proto->p_proto)) < 0)
    { printf ("failed: icmp socket()\n"); exit (0); }
  if ((channel_out = socket (AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0)
    { printf ("failed: raw socket()\n"); exit (0); }
  sequence = 0x00;
  lastport = 0xffff;
  setuid (getuid()); 
/* Ha! A Couple of non-POSIX calls! */
/*  seteuid (getuid());  */
/*  setruid (getuid()); */
  /** Hence forth the program is running as the envoker's UID **/
  current_x = 20; current_y = 20;
  packet_size = sizeof (packet_type) + pad_size;
  if ((packet_out = (packet_type *) malloc (packet_size)) == (packet_type *) NULL)
    { printf ("malloc: unable to allocate\n"); exit (0); }
  bzero((char *) packet_out,packet_size);
  fcntl (sock,F_SETFL,FNDELAY);
#ifdef SO_SNDBUF
  if (setsockopt (channel_out,SOL_SOCKET,SO_SNDBUF,
                  (char *)&packet_size,sizeof (packet_size)) < 0)
    { printf ("SO_SNDBUF: problem\n"); exit (0); }
#endif
  network = (node_type *) NULL;
  base_x = 50; base_y = 50;
  open_display (argc,argv);
  idle = TRUE;
  moving_machine = MOVING_FALSE;
  strcpy (Buttons [BUTTON_ERROR].label,"");
  strcpy (Buttons [BUTTON_NAME].label,"");
  for (;Buttons [BUTTON_QUIT].flag != TRUE;)
    {
      if ((Buttons [BUTTON_ADD].flag == TRUE) &&
          (Buttons [BUTTON_ACTIVE].flag == FALSE))
        {
          Buttons [BUTTON_ACTIVE].flag = TRUE;
          drawing (Buttons [BUTTON_ACTIVE]);
          fire_packet (time_to_live);
          time_to_live++;
        }
      if (Buttons [BUTTON_ACTIVE].flag == TRUE)
        if (catch_packet () == TRUE)
          {
            Buttons [BUTTON_ACTIVE].flag = FALSE;
            drawing (Buttons [BUTTON_ACTIVE]);
            if (type_of_packet () == TRUE)
              {
                lastport = 0xffff;
                Buttons [BUTTON_ADD].flag = FALSE;
                drawing (Buttons [BUTTON_ADD]);
                write_error ("Trace Complete");
                add_node (ret_name);
                list = (list_type *) NULL;
              }
            else
              if (strlen (ret_name) != 0)
                add_node (ret_name);
          }
      report.type = FocusIn;  /*** Set to Dummy Value ***/
      if (XPending (display) != 0)
        {
          XNextEvent (display,&report);
          idle = FALSE;
        }
      else
        idle = TRUE;
      switch (report.type) {
        case (ButtonPress):
          button = any_buttons (report);
          switch (report.xbutton.button) {
            case (Button1):
              check_button (button);
              break;
            case (Button2):
              if (report.xbutton.window == mapwindow)
                {
                  if (((report.xbutton.x >= base_x - 5) &&
                       (report.xbutton.x <= base_x + 5)) &&
                      ((report.xbutton.y >= base_y - 5) &&
                       (report.xbutton.y <= base_y + 5)))
                    {
                      moving_machine = MOVING_BASE;
                      highlight_node ((node_type *) NULL);
                      XDefineCursor (display,mapwindow,mover);
                    }
                  else
                    if ((node_move = touching_node (report.xbutton.x,
                            report.xbutton.y)) != (node_type *) NULL)
                      {
                        moving_machine = MOVING_NODE;
                        highlight_node (node_move);
                        XDefineCursor (display,mapwindow,smover);
                      }
                }
              break;
            default: break;
          }
          break;
        case (MotionNotify):
          /** Assumed that only mapwindow creates these ... **/
          tempnode = touching_node (report.xmotion.x,report.xmotion.y);
          if (tempnode != (node_type *) NULL)
            {
              strcpy (Buttons [BUTTON_SHOW].label,tempnode->name);
              drawing (Buttons [BUTTON_SHOW]);
            }
          break;
        case (ButtonRelease):
          /** Assumed that only mapwindow creates these ... **/
          if (moving_machine == MOVING_BASE)
            move_node ((node_type *) NULL,report.xbutton.x,report.xbutton.y);
          if (moving_machine == MOVING_NODE)
            move_node (node_move,report.xbutton.x,report.xbutton.y);
          moving_machine = MOVING_FALSE;
          XDefineCursor (display,mapwindow,pointer);
          break;
        case (KeyPress):
          /** Assumed that only 'window' creates these ... **/
          if (check_key (report,BUTTON_NAME) == TRUE)
            {
              if (Buttons [BUTTON_ADD].flag == TRUE)
                write_error ("Must Stop Trace First");
              else
                if (strlen (Buttons [BUTTON_NAME].label) == 0)
                  write_error ("Hostname Required");
                else
                  {
                    Buttons [BUTTON_ADD].flag = TRUE;
                    lastport = 0xffff;
                    drawing (Buttons [BUTTON_ADD]);
                    trace ();
                  }
            }
          else
            stop_trace ();
          break;
        case (LeaveNotify):
          /** Assumed that only mapwindow creates these **/
          if (moving_machine == MOVING_BASE)
            {
              moving_machine = MOVING_FALSE;
              move_node ((node_type *) NULL,base_x,base_y);
              XDefineCursor (display,mapwindow,pointer);
            }
          break;
        case (Expose):
          /** Eat up all the extra expose events and redraw window **/
          if (XPending (display) != 0)
            for (;XCheckWindowEvent (display,window,ExposureMask,&tempe) == True;);
          update_window (report.xexpose.window);
          break;
        default: break;
      }
      usleep (200);
    }
  usleep (800);
  XCloseDisplay (display);
  exit (1);
}
